#ifndef PA_H
#define PA_H

//This guy always wants to be at the very top of everything whenever 
//you are using STL with MSVC
#pragma warning(disable: 4786) 

#include <iostream>	// VIVA STL!!
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>


#include <ifl/iflFile.h> // VIVA SGI!
#include <ifl/iflTypeNames.h> 

#include <sven.h> //gen3 header

#define RECT_DEBUG 1

namespace sgg_tools {
using namespace sven;
using namespace std_ex;

struct ColorKey {

	//ColorKey::invalid gets thrown if we try to load a ColorKey file that failed to
	//conform to the basic formatting requirements
	struct invalid {};

	struct color_name_pair { 
		rgb color; string name; 
		color_name_pair(rgb color, string name) : color(color), name(name) {}
	};

	rgb background;
	vector<string> names;
	vector<rgb> colors;
	void load(const char * fname) throw(invalid);
	ColorKey() {}
	ColorKey(const char * fname) { load(fname); }
};

//a PickingAreas struct holds picking area info.
//the print function creates a .pa file that holds said info.
//the load function intilizes us from one of the aforementioned .pa files.
class PickingAreas {

	int force_name(rgb &color, ColorKey &k,vector<ColorKey::color_name_pair> &missed);
	static void find_rects(vector<Int_Rect> &rects, unsigned char * data, iflSize dims, ColorKey &key);


public:
	struct image_to_pa_error {};

	//thrown if ifl's basic open function fails
	struct open_error : public image_to_pa_error {};

	//thrown if ifl thinks that the image's dims !=3 or 4
	struct wrong_dims : public image_to_pa_error { int d; wrong_dims(int d) : d(d) {}};

	//thrown if ifl can't convert the image to byte data
	struct byte_convert_failed : public image_to_pa_error {};

	//thrown if my rectangle finding algorithms finds too many rectangles.
	//(this will almost certainly happen if the image contains rectangles
	// with anti-aliased images, or if I can't split the image into a couple of
	// of rectangular regions.)
	struct too_many_rects : public image_to_pa_error {};

	//thrown if the algorithm fails to convert the polygons it finds
	//to rectangles (if sven was less lazy, the algorithms would correctly
	//handle any polygon shaped area).
	struct sven_is_lazy : public image_to_pa_error {};

	struct pa_to_image_error {};
	//thrown if ifl's basic image create function fails
	struct image_create_error  : public pa_to_image_error {};

	//thrown is ifl's setPage function fails (during image creation)
	struct image_setPage_error : public pa_to_image_error {};


	//the color key dosen't specify a needed color / name pair
	//(the unmatched name is contained in the string)
	struct bad_color_file : public pa_to_image_error, public string {
		bad_color_file(string s) : string(s) {}
	};

	//how many rectangles can we find before we break?
	static const int RECT_MAX;

	vector<Float2_PolygonList> polys;
	vector<string> names;



	//standard output function
	void make_file(const char * fname);

	//create a bitmap coresponding to the pa.
	//make sure that fname ~= *.bmp
	//(index failed happens if the colorkey doesn't have a color defined for every area in the PA)
	void make_image(const char * fname,ColorKey &key,int width=640, int height=480)
		throw(pa_to_image_error); 

	//you have the option of appending the contents of the file to what is already in the PA
	void load_file(const char * fname,bool append=0);
	void load_string(string & file, bool append=0);

	void load_image(const char * image_name, ColorKey & key=ColorKey(), 
		vector<ColorKey::color_name_pair> &missed=vector<ColorKey::color_name_pair>(), 
		bool append=0) throw(image_to_pa_error);

	PickingAreas(const char * image_name, ColorKey &key,vector<ColorKey::color_name_pair> &missed=vector<ColorKey::color_name_pair>()) 
		{ load_image(image_name,key,missed); }
	PickingAreas(const char * file_name) { load_file(file_name); }
};

}
#endif