#include "std_precomp.h"
#include <sven.h>
#include "stam.h"

using namespace sven;

namespace Stam {
extern float * u, * v;
extern float dt;
extern bool * blocked;
extern int N;

inline int IX(int i, int j) { return ((i)+(N+2)*(j)); }

}

//triangle needs this
#define ANSI_DECLARATORS
#define REAL double
#include "triangle.h"

double get_halton(int index,int prime) { 
		index++;
        double remainder;
        double output = 0.0;
        double fraction = 1.0 / (double)prime;
        int N1 = 0;
        int copyOfIndex = index;

        while(copyOfIndex > 0) {
            N1 = (copyOfIndex / prime);
            remainder = copyOfIndex % prime;
            output += fraction * remainder;
            copyOfIndex = (int)(copyOfIndex / prime);
            fraction /= (double)prime;
        } 
        return output;
}



namespace npr {

using namespace sven;

float max_area=.001f;
float min_dist=.0002f;
float max_brush_len=0.05f;
float min_brush_len=0.01f;
float brush_vel_ratio=0.2f;

int delta_color_max=15;
int delta_color_min=-15;
float scale_intensity_min=.85f;
float scale_intensity_max=1.15f;
float random_width_min=3;
float random_width_max=3;


typedef void (BColor_fun)();

//lame hack
void white_brush() { 
	glPushAttrib(GL_COLOR_BUFFER_BIT); 
	glClearColor(1,1,1,1); 
	glClear(GL_COLOR_BUFFER_BIT);
	glPopAttrib(); 
}

void red_brush()   { 
	glPushAttrib(GL_COLOR_BUFFER_BIT); 
	glClearColor(1,0,0,1); 
	glClear(GL_COLOR_BUFFER_BIT);
	glPopAttrib(); 
}
BColor_fun * bcolor_funcs[] = { &white_brush, &red_brush, &Stam::draw_density };
BColor_fun * brush_color = bcolor_funcs[0];


///////////////////////////////////////////
//doing things this way really slows down simulation time ...  
//But, I can set up decent paramaters with
//while delany trianglulations are turned off, so this doesn't really
//need to be fast. still, the brushes can be rendered about 5 times
//faster if you assumed that the brush data was just whatever was in the
//main rendering buffer at the time.
unsigned char * getBrushData() {	
	GLint	viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	unsigned char * data = new unsigned char[viewport[2]*viewport[3]*3];

	//save the current render buffer in the accumulation buffer
	glClearAccum(0,0,0,0);
	glClear(GL_ACCUM_BUFFER_BIT);
	glAccum(GL_ACCUM, 1.f);

	//generate the brush color rendering
	glClear(GL_COLOR_BUFFER_BIT);
	brush_color();

	glFlush();
	glReadPixels(viewport[0],viewport[1],viewport[2],viewport[3], 
		GL_RGB ,    GL_UNSIGNED_BYTE,	data );

	//get back the saved rendering
	glClear(GL_COLOR_BUFFER_BIT);
	glAccum(GL_RETURN, 1.f);

	return data;
}


struct float2_rgba : public float2, public rgba {
	int w;
	float2_rgba(float2 f2) : float2(f2) {} //used by the boundry points
	float2_rgba(float2 f2, rgba c, int w) : float2(f2), rgba(c), w(w) {}
};

template <int c>
int clamp(int i) {
	if(i>c) return c;
	if(i<0) return 0;
	return i;
}

vector<float2_rgba> generate_points(int n, int p1=2, int p2=3) {
	GLint	viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	unsigned char * data = getBrushData();
	
	vector <float2_rgba> rval;
	for(int i=0;i<n;i++) {
			float x=get_halton(i,p1), y=get_halton(i,p2);
			int px=x*viewport[2];
			int py=y*viewport[3];
			if (px<0) px=0; if (px>=viewport[2]) px=viewport[2]-1; 
			if (py<0) py=0; if (py>=viewport[3]) py=viewport[3]-1; 
			int d=px+py*viewport[2];
		

	
		float c_scale=scale_intensity_min + randFloat(scale_intensity_max-scale_intensity_min);
		int dr=delta_color_min+randInt(delta_color_max-delta_color_min);
		int dg=delta_color_min+randInt(delta_color_max-delta_color_min);
		int db=delta_color_min+randInt(delta_color_max-delta_color_min);

		int r=clamp<255>(c_scale*(data[3*d+0]+dr));
		int g=clamp<255>(c_scale*(data[3*d+1]+dg));
		int b=clamp<255>(c_scale*(data[3*d+2]+db));

		rval.push_back(
			float2_rgba(
				float2(x,y),
				rgba(r,g,b,255),
				random_width_min+randFloat(random_width_max-random_width_min)
			));
	}

	delete [] data;

	return rval;
}



//vector<Stroke *> strokes;
vector<float2_rgba> tripoints; //(has the boundry points included)
vector<int3> triangles;


vector<int3> generate_triangles() {
	triangulateio in, mid;

	tripoints.push_back(float2(0,0));
	tripoints.push_back(float2(1,0));
	tripoints.push_back(float2(0,1));
	tripoints.push_back(float2(1,1));

	in.numberofpoints=tripoints.size();

	in.pointlist=(REAL *)malloc(2*(tripoints.size())*sizeof(REAL));
	int i;
	for(i=0;i<tripoints.size();i++) {
		in.pointlist[i*2]=tripoints[i].x;
		in.pointlist[i*2+1]=tripoints[i].y;
	}


	//and this is why God made constructors...
	in.numberofpointattributes=0;
	in.numberoftriangleattributes=0;
	in.pointmarkerlist = (int *) NULL;



	//mid.pointlist = (REAL*)(NULL);            // Not needed if -N switch used. 
	//mid.pointmarkerlist = (int *) NULL; // Not needed if -N or -B switch used. 
	mid.trianglelist = (int *) NULL;          // Not needed if -E switch used. 
 
	triangulate("NzQ", &in, &mid, NULL);

	vector<int3> rval;
	{for(int i=0;i<mid.numberoftriangles;i++) {
		rval.push_back(int3(mid.trianglelist+3*i));
	}}

	free(mid.trianglelist);
	free(in.pointlist);
	free(in.pointmarkerlist);

	return rval;
}

inline void glVertex(float2 p) {
	glVertex2f(p.x,p.y);
}

inline void wireTriangle(int3 t) {
	glVertex(tripoints[t[0]]);
	glVertex(tripoints[t[1]]);
	glVertex(tripoints[t[2]]);
}

void display() {
	using namespace display;
	
	glPushAttrib(GL_CURRENT_BIT | GL_POLYGON_BIT | GL_ENABLE_BIT | GL_LINE_BIT );

	glColor3f(1, 1, 1);

	if(show_triangles) {
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glBegin(GL_TRIANGLES);
		for(int i=0;i<triangles.size();i++) {
			wireTriangle(triangles[i]);
		}
		
		glEnd();
	}

	glColor3f(0, 0, 1);
	
 
	if(show_points) {
		glPointSize(3.0);
		glBegin(GL_POINTS);
		for(int i=0;i<tripoints.size();i++) {
			glVertex2f(tripoints[i].x,tripoints[i].y);
		}	
		glEnd();
	}

	if(show_brushes && tripoints.size()) {


		
		glEnable(GL_LINE_SMOOTH);
		
		for(vector<float2_rgba>::iterator i=tripoints.begin();i!=tripoints.end()-4;i++) {
			using namespace Stam;
			int i0,i1,j0,j1; 
			float x=i->x*N,y=i->y*N,vx, vy, s0,s1,t0,t1;

			//enforce boundry contraints
			if (x<0.5f) x=0.5f; if (x>N+0.5f) x=N+0.5f; 
			if (y<0.5f) y=0.5f; if (y>N+0.5f) y=N+0.5f; 

			i0=(int)x; i1=i0+1;
			j0=(int)y; j1=j0+1;

			if( blocked[IX(i0,j0)] && 
				blocked[IX(i1,j0)] && 
				blocked[IX(i0,j1)] && 
				blocked[IX(i1,j1)]) continue;

			s1 = x-i0; s0 = 1-s1; t1 = y-j0; t0 = 1-t1;

			vx = (s0*(t0*u[IX(i0,j0)]+t1*u[IX(i0,j1)])+
				   s1*(t0*u[IX(i1,j0)]+t1*u[IX(i1,j1)]));
					
			vy = (s0*(t0*v[IX(i0,j0)]+t1*v[IX(i0,j1)])+
				   s1*(t0*v[IX(i1,j0)]+t1*v[IX(i1,j1)]));

			float bx=vx*brush_vel_ratio, 
				  by=vy*brush_vel_ratio;

			float m2=bx*bx+by*by;
			if(m2>max_brush_len*max_brush_len) {
				float m=sqrt(m2);
				float r= max_brush_len/m;
				bx=bx*r;
				by=by*r;
			}
			if(m2<min_brush_len*min_brush_len) {
				if(!m2) {
					bx=min_brush_len;
				}
				else {
				float m=sqrt(m2);
				float r= min_brush_len/m;
				bx=bx*r;
				by=by*r;
				}
			}

			glLineWidth(i->w);
			//ostringstream out;
			//out << i->w;
			//win_tools::fyi_msg(out.str().c_str());
			glBegin(GL_LINES);
			glColor4f(i->r/255.,i->g/255.,i->b/255.,1);
			glVertex2f(i->x-bx,i->y-by);
			glVertex2f(i->x+bx,i->y+by);
			glEnd();
		}
		
	}

	glPopAttrib();
}

inline float tri_area(const int3 &triangle) {
	float ux = tripoints[triangle[0]].x-tripoints[triangle[1]].x;
	float vx = tripoints[triangle[0]].x-tripoints[triangle[2]].x;
	float uy = tripoints[triangle[0]].y-tripoints[triangle[1]].y;
	float vy = tripoints[triangle[0]].y-tripoints[triangle[2]].y;
	return sven::abs((ux*vy-vx*uy)*.5);
}

inline float sq(float f) { return f*f; }

//taken from Shewchuk's lecture notes on geometric stabilty
inline float2 circumcenter(const int3 &triangle, const float A) {
	float ax = tripoints[triangle[0]].x;
	float ay = tripoints[triangle[0]].y;
	float bx = tripoints[triangle[1]].x;
	float by = tripoints[triangle[1]].y;
	float cx = tripoints[triangle[2]].x;
	float cy = tripoints[triangle[2]].y;
	return float2(
		  cx + ( (sq(ax-cx)+sq(ay-cy)) * (by-cy) - (sq(bx-cx) + sq(by-cy))*(ay-cy) )/(4*A),
		  cy + ( (sq(bx-cx)+sq(by-cy)) * (ax-cx) - (sq(ax-cx) + sq(ay-cy))*(bx-cx) )/(4*A));
		
}


inline float mag2(const float2 &a, const float2 &b) {
	return (a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y);
}

void idle() {
	//move the points
	if(tripoints.size()) {

		//Ok kids, it's time to play "SPOT THE KLUDGE" !
		tripoints.pop_back();
		tripoints.pop_back();
		tripoints.pop_back();
		tripoints.pop_back();

		{for(vector<float2_rgba>::iterator i=tripoints.begin();i!=tripoints.end();i++) {
		    using namespace Stam;
			int i0,i1,j0,j1; float x=i->x*N,y=i->y*N,vx, vy, s0,s1,t0,t1,tleft,tnext,dt0=dt*N,t;
			i0=(int)x; i1=i0+1;
			j0=(int)y; j1=j0+1;
			s1 = x-i0; s0 = 1-s1; t1 = y-j0; t0 = 1-t1;
			//(all this is stolen from the advection code)

		tleft=dt0;		

		while(tleft>smallf) {

			//enforce boundry contraints
			if (x<0.5f) x=0.5f; if (x>N+0.5f) x=N+0.5f; 
			if (y<0.5f) y=0.5f; if (y>N+0.5f) y=N+0.5f; 


			i0=(int)x; i1=i0+1;
			j0=(int)y; j1=j0+1;
			s1 = x-i0; s0 = 1-s1; t1 = y-j0; t0 = 1-t1;

			vx = (s0*(t0*u[IX(i0,j0)]+t1*u[IX(i0,j1)])+
				   s1*(t0*u[IX(i1,j0)]+t1*u[IX(i1,j1)]));
					
			vy = (s0*(t0*v[IX(i0,j0)]+t1*v[IX(i0,j1)])+
				   s1*(t0*v[IX(i1,j0)]+t1*v[IX(i1,j1)]));


			float speed2=vx*vx+vy*vy; 
			if(speed2>smallf) tnext=.5/sqrt(speed2);
			else tnext=tleft;

			t=tnext > tleft ? tleft : tnext;
			tleft-=t;


			x+=t*vx;
			y+=t*vy;
		}


			i->x=x/N; i->y=y/N;
		}}

		//it turns out that deleting points which are inside of 
		//boundries isn't really a good idea.
		//The max and min contraints really do enough between them
		//to keep points from getting overly clustered around or
		//inside of boundries.
		/*for(int i=0;i<tripoints.size();) {
			int i0,j0,i1,j1;
			using namespace Stam;
			float x=tripoints[i].x*N,y=tripoints[i].y*N;
			i0=(int)x; i1=i0+1;
			j0=(int)y; j1=j0+1;
			if(blocked[IX(i0,j0)] || blocked[IX(i1,j1)] ) tripoints.erase(tripoints.begin()+i);
			else i++;
		}*/



		{

	GLint	viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	unsigned char * data = getBrushData();

		//add new points
		vector<float2_rgba> new_points;
		for(int i=0;i<triangles.size();i++) {
			float A=tri_area(triangles[i]);
			if(A>max_area) {
			float2 f2 = circumcenter(triangles[i],A);
			float x=f2.x,y=f2.y;

			int px=x*viewport[2];
			int py=y*viewport[3];

			if (px<0) px=0; if (px>=viewport[2]) px=viewport[2]-1; 
			if (py<0) py=0; if (py>=viewport[3]) py=viewport[3]-1; 
	
			int d=px+py*viewport[2];
		
			float c_scale=scale_intensity_min + randFloat(scale_intensity_max-scale_intensity_min);
			int dr=delta_color_min+randInt(delta_color_max-delta_color_min);
			int dg=delta_color_min+randInt(delta_color_max-delta_color_min);
			int db=delta_color_min+randInt(delta_color_max-delta_color_min);

			int r=clamp<255>(c_scale*(data[3*d+0]+dr));
			int g=clamp<255>(c_scale*(data[3*d+1]+dg));
			int b=clamp<255>(c_scale*(data[3*d+2]+db));

			new_points.push_back(
				float2_rgba(
					float2(x,y),
					rgba(r,g,b,255),
					random_width_min+randFloat(random_width_max-random_width_min)
				));


			}
		}

		delete [] data;

		for(i=0;i<new_points.size();i++) {
			tripoints.insert(tripoints.begin()+randInt(tripoints.size()),new_points[i]);
		}

		}


		for(int i=0;i<tripoints.size()-1;) {
			for(int j=i+1;j<tripoints.size();j++) {
				if(mag2(tripoints[i],tripoints[j]) < min_dist) {
					tripoints.erase(tripoints.begin()+i);
					break;
				}
			}
			if(j==tripoints.size()) i++;
		}

		//(note that this pushes back on the extra points that we just popped off)
		triangles=generate_triangles();
	}

}

void clear_points() {
	tripoints.clear();
	triangles.clear();
}

}


void MyFrame::NPR_InitPoints(wxCommandEvent& event) {
	long res=800;

	if(res!=-1){
		npr::tripoints=npr::generate_points(res);
		npr::triangles=npr::generate_triangles();
	}
}
