#include "nehe.h"

namespace gl_ex {

using namespace sven;
using namespace std;

//Sven hacked Nehe code at it's best.  The is a geneneralized version
//of Nehe's origional picking code.  
//x and y are the window coordinates to pick under 
//(usually these are just the mouse coordinates).
//
//drawItems is a function that draws everything that you want to be pickable.
//before each pickable section of the drawing call a glLoadName(n) for n!=-1.
//The return value of this function will be -1 if your window coordinates
//are not ontop of anything, and n if the pickable section with glLoadName(n)
//was ontop.  Object selection made easy.
//
//mm is the modelview matrix that you want to be used.  If you like you
//can not specify this paramater, in which case the current matrix will be used.
int glSelect(GLdouble x, GLdouble y, void_fun drawItems, GLfloat *mm) {
	
	//handle the null mm situation
	GLfloat cmm[16];
	if(!mm) {
		glGetFloatv(GL_MODELVIEW_MATRIX,cmm);
		mm=cmm;
	}


	GLuint	buffer[512];							// Set Up A Selection Buffer
	int hits=0;

	glMatrixMode(GL_MODELVIEW);						// Select The Modelview Matrix
	glPushMatrix();
	glLoadIdentity();
	glMultMatrixf(mm);

	// The Size Of The Viewport. [0] Is <x>, [1] Is <y>, [2] Is <length>, [3] Is <width>
	GLint	viewport[4];

	// This Sets The Array <viewport> To The Size And Location Of The Screen Relative To The Window
	glGetIntegerv(GL_VIEWPORT, viewport);
	glSelectBuffer(512, buffer);						// Tell OpenGL To Use Our Array For Selection
	glRenderMode(GL_SELECT);	
	glInitNames();								// Initializes The Name Stack
	glPushName(0);
	glMatrixMode(GL_PROJECTION);						// Selects The Projection Matrix
	glPushMatrix();								// Push The Projection Matrix
	glLoadIdentity();							// Resets The Matrix	
	gluPickMatrix(x, viewport[3]-y, 15.0f, 15.0f, viewport);

	//this might need some editting....
	using namespace nehe;
	gluPerspective(PERSP_ANGLE,(GLfloat) (viewport[2]-viewport[0])/(GLfloat) (viewport[3]-viewport[1]),PERSP_NEAR,PERSP_FAR);

	glMatrixMode(GL_MODELVIEW);						// Select The Modelview Matrix
	drawItems();
	glPopMatrix();

	int choose = -1;

	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glPopMatrix();								// Pop The Projection Matrix
	glMatrixMode(GL_MODELVIEW);						// Select The Modelview Matrix

	hits=glRenderMode(GL_RENDER);						// Switch To Render Mode, Find Out How Many

	if (hits > 0)								// If There Were More Than 0 Hits
	{
		choose = buffer[3];					// Make Our Selection The First Object
		int	depth = buffer[1];					// Store How Far Away It Is

		//fyi_msg("you got a hit");

		for (int loop = 1; loop < hits; loop++)				// Loop Through All The Detected Hits
		{
			// If This Object Is Closer To Us Than The One We Have Selected
			if (buffer[loop*4+1] < GLuint(depth))
			{
				choose = buffer[loop*4+3];			// Select The Closer Object
				depth = buffer[loop*4+1];			// Store How Far Away It Is
			}       
		}

	}

	return choose;
}

//Ok, so in opengl, you specify the raster position in object
//coordinates, but when you ask the API to give you the raster position,
//they use object coordinates.  This can be annoying.  Thus, this function.
//If gives you back the raster position, measured in object coordinates. 
void glRasterObjPos(GLdouble * o_pos) {
	GLdouble r_pos[4];
	glGetDoublev(GL_CURRENT_RASTER_POSITION,r_pos);
	GLdouble modelMatrix[16];
	glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
	GLdouble projMatrix[16];   
	glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT,viewport);
	gluUnProject(r_pos[0], r_pos[1], r_pos[2],                                   
				 modelMatrix, projMatrix, viewport,         
				 o_pos+0,o_pos+1,o_pos+2);

}

void glPrintMatrix() {
	float m[16]; 
	glGetFloatv(GL_MODELVIEW_MATRIX,m);

	using namespace nehe;

	glPrintf_px(TextFormat_px(100,100),"%f %f %f %f\n%f %f %f %f\n%f %f %f %f\n%f %f %f %f",
		m[0],m[4],m[8],m[12],
		m[1],m[5],m[9],m[13],
		m[2],m[6],m[10],m[14],
		m[3],m[7],m[11],m[15]);
}


void print4x4(const char * fname, float * m) {
	ofstream out(fname);
	
	for(int c=0; c<4; c++) {
		for(int r=0; r<4; r++) {
			out << m[c+r*4] << " ";
		}
		out << endl;
	}

	out.close();
}


void invCTransformMatrix(float3 i,float3 j,float3 k, float * r) {

	r[0]=i[0]; r[1]=i[1]; r[2]=i[2]; r[3]=0,
	r[4]=j[0]; r[5]=j[1]; r[6]=j[2]; r[7]=0,
	r[8]=k[0]; r[9]=k[1]; r[10]=k[2];r[11]=0,
	r[12]=0;   r[13]=0;	  r[14]=0;	 r[15]=1;
}

void cordTransformMatrix(float3 i,float3 j,float3 k, float * r) {

	r[0]=i[0]; r[1]=j[0]; r[2]=k[0]; r[3]=0,
	r[4]=i[1]; r[5]=j[1]; r[6]=k[1]; r[7]=0,
	r[8]=i[2]; r[9]=j[2]; r[10]=k[2];r[11]=0,
	r[12]=0;   r[13]=0;	  r[14]=0;	 r[15]=1;
}

//for some reason, i was not in the mood to write a for loop.  shrug.
void copy4x4(float * m1, float * m2) {
	m2[0]=m1[0];m2[1]=m1[1];m2[2]=m1[2];m2[3]=m1[3];m2[4]=m1[4];m2[5]=m1[5];m2[6]=m1[6];m2[7]=m1[7];m2[8]=m1[8];m2[9]=m1[9];m2[10]=m1[10];m2[11]=m1[11];m2[12]=m1[12];m2[13]=m1[13];m2[14]=m1[14];m2[15]=m1[15];
}

//rather untested, but i think it should work
void mult4x4(float * m1, float * m2, float * rval) {
	float sum;
	for(int r=0;r<4;r++) {
		for(int c=0;c<4; c++) {
			sum=0;
			for(int q=0;q<4;q++) sum+=m1[4*r+q]*m2[q*4+c];
			rval[r*4+c]=sum;
		}
	}	
}

void identity4x4(float * m) {
	m[0]=1;  m[1]=0;  m[2]=0;  m[3]=0;
	m[4]=0;  m[5]=1;  m[6]=0;  m[7]=0;
	m[8]=0;  m[9]=0;  m[10]=1; m[11]=0;
	m[12]=0; m[13]=0; m[14]=0; m[15]=1;
}

void cross(float * v1, float * v2, float * rval) {
	rval[0]=v1[1]*v2[2]-v2[1]*v1[2];
	rval[1]=v2[0]*v1[2]-v1[0]*v2[2];
	rval[2]=v1[0]*v2[1]-v2[0]*v1[1];
}

void dif(float * v1, float * v2, float * rval) {
	rval[0]=v1[0]-v2[0];
	rval[1]=v1[1]-v2[1];
	rval[2]=v1[2]-v2[2];
}

void nCross(float * v1, float * v2, float * rval) {
	float cross[3];
	float mag;
	cross[0]=v1[1]*v2[2]-v2[1]*v1[2];
	cross[1]=v2[0]*v1[2]-v1[0]*v2[2];
	cross[2]=v1[0]*v2[1]-v2[0]*v1[1];

	mag=sqrtf(cross[0]*cross[0]+cross[1]*cross[1]+cross[2]*cross[2]);
	
	rval[0]=cross[0]/mag;
	rval[1]=cross[1]/mag;
	rval[2]=cross[2]/mag;
}

float3 operator *(float * a, float3 b) {
	float3 r(a[0]*b[0]+a[4]*b[1]+a[8]*b[2]+a[12],
			a[1]*b[0]+a[5]*b[1]+a[9]*b[2]+a[13],
			a[2]*b[0]+a[6]*b[1]+a[10]*b[2]+a[14]);
	
	return r;
}	

void translateMatrix(float x, float y, float z, float * rval) {
	identity4x4(rval);
	rval[12]=x;
	rval[13]=y;
	rval[14]=z;
}

void tramslateMatrix(float3 t, float * rval) {
	identity4x4(rval);
	rval[12]=t[0];
	rval[13]=t[1];
	rval[14]=t[2];
}

void invCordTransform(float3 i,float3 j,float3 k) {
	float m[16] = {
		i[0], i[1], i[2], 0,
		j[0], j[1], j[2], 0,
		k[0], k[1], k[2], 0,
		0,	  0,    0,    1
	};
	glMultMatrixf(m);
}

void cordTransform(float3 i,float3 j,float3 k) {
	float m[16] = {
		i[0], j[0], k[0], 0,
		i[1], j[1], k[1], 0,
		i[2], j[2], k[2], 0,
		0,	  0,    0,    1
	};
  
  glMultMatrixf(m);
}


vector<float4> matDifStack;

void glPushMaterialDif() {
	float4 ld;
	glGetMaterialfv(GL_FRONT, GL_DIFFUSE, ld.v);
	matDifStack.push_back(ld);
}

void glPopMaterialDif() {
	float4 ld;
	ld=matDifStack.back();
	matDifStack.pop_back();
	glMaterialfv(GL_FRONT, GL_DIFFUSE,ld.v);
}

void glMaterialDif(float r, float g, float b, float w) {
	GLfloat ld[] = {r,g,b,w};
	glMaterialfv(GL_FRONT, GL_DIFFUSE,ld);
}


}